#include <string>
#include <vector>
#include <fstream>
#include <filesystem>
#include <cstring>
#include <zlib.h>
#include <sstream>
#include <span>

#include "tar_to_stream.hpp"
#include "ArchiveManager.hpp"

namespace fs = std::filesystem;

namespace {
constexpr size_t TAR_BLOCK_SIZE = 512;

// --- Minimal tar extraction logic for unpacking and config extraction ---
struct TarHeader {
    char name[100];
    char mode[8];
    char uid[8];
    char gid[8];
    char size[12];
    char mtime[12];
    char chksum[8];
    char typeflag;
    char linkname[100];
    char magic[6];
    char version[2];
    char uname[32];
    char gname[32];
    char devmajor[8];
    char devminor[8];
    char prefix[155];
    char pad[12];
};

size_t tarFileSize(const TarHeader* hdr) {
    return std::strtol(hdr->size, nullptr, 8);
}

std::string tarFileName(const TarHeader* hdr) {
    std::string name(hdr->name);
    if (hdr->prefix[0]) {
        std::string prefix(hdr->prefix);
        return prefix + "/" + name;
    }
    return name;
}

bool readTar(const std::vector<uint8_t>& tarData, const std::string& outDir, std::string* configJson = nullptr) {
    size_t pos = 0;
    while (pos + sizeof(TarHeader) <= tarData.size()) {
        const TarHeader* hdr = reinterpret_cast<const TarHeader*>(&tarData[pos]);
        if (hdr->name[0] == '\0') break;
        size_t filesize = tarFileSize(hdr);
        std::string filename = tarFileName(hdr);
        size_t fileStart = pos + sizeof(TarHeader);
        if (configJson && filename == "dropshell-tool-config.json") {
            *configJson = std::string(reinterpret_cast<const char*>(&tarData[fileStart]), filesize);
        } else if (!outDir.empty()) {
            fs::path outPath = fs::path(outDir) / filename;
            fs::create_directories(outPath.parent_path());
            std::ofstream ofs(outPath, std::ios::binary);
            ofs.write(reinterpret_cast<const char*>(&tarData[fileStart]), filesize);
        }
        pos = fileStart + filesize;
        pos += (TAR_BLOCK_SIZE - (filesize % TAR_BLOCK_SIZE)) % TAR_BLOCK_SIZE;
    }
    return true;
}

bool extractConfigJson(const std::vector<uint8_t>& tarData, std::string& outJson) {
    return readTar(tarData, "", &outJson);
}

bool replaceConfigJson(std::vector<uint8_t>& tarData, const std::string& json) {
    // Remove old config, add new one at end
    std::vector<uint8_t> newTar;
    size_t pos = 0;
    bool replaced = false;
    while (pos + sizeof(TarHeader) <= tarData.size()) {
        const TarHeader* hdr = reinterpret_cast<const TarHeader*>(&tarData[pos]);
        if (hdr->name[0] == '\0') break;
        size_t filesize = tarFileSize(hdr);
        std::string filename = tarFileName(hdr);
        size_t fileStart = pos + sizeof(TarHeader);
        if (filename != "dropshell-tool-config.json") {
            newTar.insert(newTar.end(), &tarData[pos], &tarData[fileStart + filesize]);
            size_t pad = (TAR_BLOCK_SIZE - (filesize % TAR_BLOCK_SIZE)) % TAR_BLOCK_SIZE;
            newTar.insert(newTar.end(), pad, 0);
        } else {
            replaced = true;
        }
        pos = fileStart + filesize;
        pos += (TAR_BLOCK_SIZE - (filesize % TAR_BLOCK_SIZE)) % TAR_BLOCK_SIZE;
    }
    // Add new config
    // Use tar_to_stream to add the config file
    std::ostringstream oss(std::ios::binary);
    oss.write(reinterpret_cast<const char*>(newTar.data()), newTar.size());
    std::string uname = "root";
    std::string gname = "root";
    tar_to_stream_properties props = {
        "dropshell-tool-config.json",
        std::as_bytes(std::span(json.data(), json.size())),
        0u, "644", 0u, 0u, uname, gname
    };
    tar_to_stream(oss, std::move(props));
    tar_to_stream_tail(oss);
    std::string tarStr = oss.str();
    tarData.assign(tarStr.begin(), tarStr.end());
    return true;
}

// --- zlib helpers ---
bool gzipCompress(const std::string& in, std::vector<uint8_t>& out) {
    z_stream strm{};
    deflateInit2(&strm, Z_BEST_COMPRESSION, Z_DEFLATED, 15 + 16, 8, Z_DEFAULT_STRATEGY);
    out.resize(compressBound(in.size()));
    strm.next_in = reinterpret_cast<Bytef*>(const_cast<char*>(in.data()));
    strm.avail_in = in.size();
    strm.next_out = out.data();
    strm.avail_out = out.size();
    int ret = deflate(&strm, Z_FINISH);
    if (ret != Z_STREAM_END) { deflateEnd(&strm); return false; }
    out.resize(strm.total_out);
    deflateEnd(&strm);
    return true;
}

bool gzipDecompress(const std::string& inPath, std::vector<uint8_t>& out) {
    std::ifstream ifs(inPath, std::ios::binary);
    if (!ifs) return false;
    ifs.seekg(0, std::ios::end);
    size_t insize = ifs.tellg();
    ifs.seekg(0, std::ios::beg);
    std::vector<uint8_t> inbuf(insize);
    ifs.read(reinterpret_cast<char*>(inbuf.data()), insize);
    z_stream strm{};
    inflateInit2(&strm, 15 + 16);
    out.resize(insize * 10); // crude guess
    strm.next_in = inbuf.data();
    strm.avail_in = inbuf.size();
    strm.next_out = out.data();
    strm.avail_out = out.size();
    int ret = inflate(&strm, Z_FINISH);
    if (ret != Z_STREAM_END) { inflateEnd(&strm); return false; }
    out.resize(strm.total_out);
    inflateEnd(&strm);
    return true;
}

bool gzipDecompressToTar(const std::string& inPath, std::vector<uint8_t>& tarData) {
    return gzipDecompress(inPath, tarData);
}

bool gzipCompressToFile(const std::string& tarData, const std::string& outPath) {
    std::vector<uint8_t> gz;
    if (!gzipCompress(tarData, gz)) return false;
    std::ofstream ofs(outPath, std::ios::binary);
    if (!ofs) return false;
    ofs.write(reinterpret_cast<const char*>(gz.data()), gz.size());
    return ofs.good();
}
}

ArchiveManager::ArchiveManager() {}

bool ArchiveManager::pack(const std::string& folderPath, const std::string& archivePath) {
    // Use tar_to_stream to create tar in memory
    std::ostringstream tarStream(std::ios::binary);
    for (auto& p : fs::recursive_directory_iterator(folderPath)) {
        if (!fs::is_regular_file(p)) continue;
        std::ifstream ifs(p.path(), std::ios::binary);
        if (!ifs) return false;
        std::vector<char> data((std::istreambuf_iterator<char>(ifs)), {});
        std::string uname = "root";
        std::string gname = "root";
        tar_to_stream_properties props = {
            fs::relative(p.path(), folderPath).generic_string(),
            std::as_bytes(std::span(data.data(), data.size())),
            0u, "644", 0u, 0u, uname, gname
        };
        tar_to_stream(tarStream, std::move(props));
    }
    tar_to_stream_tail(tarStream);
    std::string tarData = tarStream.str();
    return gzipCompressToFile(tarData, archivePath);
}

bool ArchiveManager::unpack(const std::string& archivePath, const std::string& outDir) {
    std::vector<uint8_t> tarData;
    if (!gzipDecompressToTar(archivePath, tarData)) return false;
    return readTar(tarData, outDir);
}

bool ArchiveManager::readConfigJson(const std::string& archivePath, std::string& outJson) {
    std::vector<uint8_t> tarData;
    if (!gzipDecompressToTar(archivePath, tarData)) return false;
    return extractConfigJson(tarData, outJson);
}

bool ArchiveManager::writeConfigJson(const std::string& archivePath, const std::string& json) {
    std::vector<uint8_t> tarData;
    if (!gzipDecompressToTar(archivePath, tarData)) return false;
    if (!replaceConfigJson(tarData, json)) return false;
    // Write back to archivePath
    std::string tarStr(reinterpret_cast<const char*>(tarData.data()), tarData.size());
    return gzipCompressToFile(tarStr, archivePath);
} 